import numpy as np
import evaluate
from datasets import load_from_disk
from transformers import (
    AutoTokenizer,
    AutoModelForTokenClassification,
    TrainingArguments,
    Trainer,
    DataCollatorForTokenClassification,
)


def main():
    model_name = "distilbert/distilbert-base-uncased"
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    dataset = load_from_disk("data/refined-bookcorpus-dataset_hf_split")
    print(dataset)
    dataset_val = dataset["test"]
    dataset_train = dataset["train"]
    label_list = ["O", "separator"]
    seqeval = evaluate.load("seqeval")

    def tokenize_and_align_labels(examples):
        tokenized_inputs = tokenizer(
            examples["tokens"], truncation=True, is_split_into_words=True
        )

        labels = []
        for i, label in enumerate(examples["ner_tags"]):
            word_ids = tokenized_inputs.word_ids(
                batch_index=i
            )  # Map tokens to their respective word.
            previous_word_idx = None
            label_ids = []
            for word_idx in word_ids:  # Set the special tokens to -100.
                if word_idx is None:
                    label_ids.append(-100)
                elif (
                    word_idx != previous_word_idx
                ):  # Only label the first token of a given word.
                    label_ids.append(label[word_idx])
                else:
                    label_ids.append(-100)
                previous_word_idx = word_idx
            labels.append(label_ids)

        tokenized_inputs["labels"] = labels

        return tokenized_inputs

    def compute_metrics(p):
        predictions, labels = p
        predictions = np.argmax(predictions, axis=2)

        true_predictions = [
            [label_list[p] for (p, l) in zip(prediction, label) if l != -100]
            for prediction, label in zip(predictions, labels)
        ]
        true_labels = [
            [label_list[l] for (p, l) in zip(prediction, label) if l != -100]
            for prediction, label in zip(predictions, labels)
        ]

        results = seqeval.compute(predictions=true_predictions, references=true_labels)
        return {
            "precision": results["overall_precision"],
            "recall": results["overall_recall"],
            "f1": results["overall_f1"],
            "accuracy": results["overall_accuracy"],
        }

    tokenized_dataset_train = dataset_train.map(tokenize_and_align_labels, batched=True)
    tokenized_dataset_val = dataset_val.map(tokenize_and_align_labels, batched=True)

    id2label = {
        0: "O",
        1: "separator",
    }
    label2id = {
        "O": 0,
        "separator": 1,
    }

    model = AutoModelForTokenClassification.from_pretrained(
        model_name, num_labels=2, id2label=id2label, label2id=label2id
    )

    data_collator = DataCollatorForTokenClassification(tokenizer=tokenizer)
    training_args = TrainingArguments(
        output_dir="bookcorpus_model",
        learning_rate=2e-5,
        per_device_train_batch_size=56,
        per_device_eval_batch_size=56,
        num_train_epochs=100,
        weight_decay=0.01,
        eval_strategy="epoch",
        save_strategy="epoch",
        load_best_model_at_end=True,
        push_to_hub=False,
    )

    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=tokenized_dataset_train,
        eval_dataset=tokenized_dataset_val,
        processing_class=tokenizer,
        data_collator=data_collator,
        compute_metrics=compute_metrics,
    )

    trainer.train()


if __name__ == "__main__":
    main()



from transformers import AutoTokenizer, AutoModelForTokenClassification, pipeline


def split_into_chunks(lst, n):
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def split_text_into_even_chunks(tokenizer, text):
    ids = tokenizer(text, truncation=False, add_special_tokens=False)
    ids = ids["input_ids"]
    ids = split_into_chunks(ids, n=tokenizer.model_max_length)
    chunks = (tokenizer.decode(chunk_ids) for chunk_ids in ids)

    return chunks


def split_into_semantic_chunks(text, ners):
    begin_index = 0

    for i, _c in enumerate(text):
        for ner in ners:
            if i == ner["end"]:
                chunk = text[begin_index : ner["end"]]
                chunk = chunk.strip()
                yield chunk

                begin_index = ner["end"]

    yield text[begin_index:]


class TextSplitter:
    def __init__(self, model_name="mirth/chonky_distilbert_uncased_1", device="cpu"):
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)

        id2label = {
            0: "O",
            1: "separator",
        }
        label2id = {
            "O": 0,
            "separator": 1,
        }

        model = AutoModelForTokenClassification.from_pretrained(
            model_name,
            num_labels=2,
            id2label=id2label,
            label2id=label2id,
        )
        model.to(device)
        self.pipe = pipeline(
            "ner",
            model=model,
            tokenizer=self.tokenizer,
            device=device,
            aggregation_strategy="simple",
        )

    def __call__(self, text):
        text_chunks = split_text_into_even_chunks(self.tokenizer, text)

        for text_chunk in text_chunks:
            output = self.pipe(text_chunk)

            yield from split_into_semantic_chunks(text_chunk, output)


if __name__ == "__main__":
    with open(
        "../../data/paul_graham_essay_no_new_line/paul_graham_essay_no_new_line.txt"
    ) as file:
        pg = file.read()

    c = TextSplitter(device="cuda")
    for sem_chunk in c(pg):
        print(sem_chunk)
        print("--")
